/*
 * Device2D.java
 *
 * Created on Piatok, 2004, oktber 15, 13:24
 */

/**
 *
 * @author  Marek Mizanin
 * color buffer, z buffer, drawing for line
 */
package graphics.device;

import graphics.math.*;

public class Device2D
{
	public int[] cBuf;			// color Buffer
//	public int[] zBuf;			// z Buffer
	public int sizeX, sizeY, size;	// width, height, size = sizeX*sizeY
	public int actual_c = 0xffffff;
	
	/** Creates a new instance of Device2D */
	public Device2D(){}
	public Device2D(int sizeX_, int sizeY_)
	{
		sizeX = sizeX_;
		sizeY = sizeY_;
		size = sizeX * sizeY;
		cBuf = new int[sizeY*sizeX];
	}
	
	public void clearColorBuffer(int color)	
	{
		for(int i=0; i<size; i++)
			cBuf[i] = color;
	}
	
	public void noiseColor()
	{
		java.util.Random rnd = new java.util.Random();
		for(int i=0; i<size; i++)
		{
			cBuf[i] = (int)rnd.nextInt(); //color;
		}
	}
	
	public void drawLineF(vi a, vi b)	// Bresenham algoritmus
	{
		int dy = b.y - a.y;
		int dx = b.x - a.x;
		int stepx = 1;
		int stepy = 1;
		int x = a.x;
		int y = a.y;

		if (dy < 0)
		{
			dy = -dy;
			stepy = -1;
		}

		if (dx < 0)
		{
			dx = -dx;
			stepx = -1;
		}

		dy <<= 1;  // dy is now 2*dy
		dx <<= 1;  // dx is now 2*dx

		if(y>=0 && y<sizeY && x>=0 && x<sizeX)
			cBuf[y*sizeX+x] = actual_c;

		if (dx > dy)
		{
			y *= sizeX;
			stepy *= sizeX;
			int fraction = dy - (dx >> 1); // same as 2*dy - dx
			while (x != b.x)
			{
				if (fraction >= 0)
				{
					y += stepy;
					fraction -= dx;  // same as fraction -= 2*dx
				}
				x += stepx;
				fraction += dy; // same as fraction -= 2*dy

				if(y>=0 && y<sizeY*sizeX && x>=0 && x<sizeX)
					cBuf[y+x] = actual_c;
			}
		}
		else
		{
			int fraction = dx - (dy >> 1);
			while (y != b.y)
			{
				if (fraction >= 0)
				{
					x += stepx;
					fraction -= dy;
				}
				y += stepy;
				fraction += dx;

				if(y>=0 && y<sizeY && x>=0 && x<sizeX)
					cBuf[y*sizeX+x] = actual_c;
			}
		}
	}
	// ac color a, bc color b
	public void drawLineG(vi a, vi b, int ac, int bc)	// Bresenham algoritmus
	{
		int dy = b.y - a.y;
		int dx = b.x - a.x;
		int stepx = 1;
		int stepy = 1;
		int x = a.x;
		int y = a.y;

		if (dy < 0)
		{
			dy = -dy;
			stepy = -1;
		}

		if (dx < 0)
		{
			dx = -dx;
			stepx = -1;
		}

		dy <<= 1;  // dy is now 2*dy
		dx <<= 1;  // dx is now 2*dx

		if(y>=0 && y<sizeY && x>=0 && x<sizeX)
			cBuf[y*sizeX+x] = ac;

		if (dx > dy)
		{
			y *= sizeX;
			stepy *= sizeX;
			int fraction = dy - (dx >> 1); // same as 2*dy - dx
			while (x != b.x)
			{
				if (fraction >= 0)
				{
					y += stepy;
					fraction -= dx;  // same as fraction -= 2*dx
				}
				x += stepx;
				fraction += dy; // same as fraction -= 2*dy

				if(y>=0 && y<sizeY*sizeX && x>=0 && x<sizeX)
					cBuf[y+x] = col.get(ac, bc, a.x, b.x, x);
			}
		}
		else
		{
			int fraction = dx - (dy >> 1);
			while (y != b.y)
			{
				if (fraction >= 0)
				{
					x += stepx;
					fraction -= dy;
				}
				y += stepy;
				fraction += dx;

				if(y>=0 && y<sizeY && x>=0 && x<sizeX)
					cBuf[y*sizeX+x] = col.get(ac, bc, a.y, b.y, y);;
			}
		}
	}
	
	public void drawLineF(vi a, vi b, int width)
	{
		drawLineF(a,b);
		if( Math.abs(b.y-a.y)>Math.abs(b.x-a.x) )
		{
			int ax=a.x-width/2;
			int bx=b.x-width/2;
			while(width>0)			
			{
				drawLineF(new vi(ax,a.y), new vi(bx,b.y));
				ax++;
				bx++;
				width--;
			}
		}
		else
		{
			int ay=a.y-width/2;
			int by=b.y-width/2;
			while(width>0)			
			{
				drawLineF(new vi(a.x,ay), new vi(b.x,by));
				ay++;
				by++;
				width--;
			}		
		}
	}
	
	public void drawLineG(vi a, vi b, int ac, int bc, int width)
	{
		drawLineG(a,b,ac,bc);
		if( Math.abs(b.y-a.y)>Math.abs(b.x-a.x) )
		{
			int ax=a.x-width/2;
			int bx=b.x-width/2;
			while(width>0)			
			{
				drawLineG(new vi(ax,a.y), new vi(bx,b.y),ac,bc);
				ax++;
				bx++;
				width--;
			}
		}
		else
		{
			int ay=a.y-width/2;
			int by=b.y-width/2;
			while(width>0)			
			{
				drawLineG(new vi(a.x,ay), new vi(b.x,by),ac,bc);
				ay++;
				by++;
				width--;
			}		
		}
	}
}

