/*
 * Device3D.java
 *
 * Created on Piatok, 2004, oktber 15, 19:34
 */

/**
 *
 * @author  Marek Mizanin
 */

package graphics.device;

import java.util.*;
import graphics.math.*;
import graphics.device.Device2D_triangle;
import graphics.device.Material;
import graphics.device.Light;

public class Device3D extends Device2D_triangle
{
	// primitives
//	final public static int GL_POINTS = 1;
	final public static int GL_LINES = 2;
	final public static int GL_LINE_STRIP = 3;
	final public static int GL_LINE_LOOP = 4;
	final public static int GL_TRIANGLES = 5;
	final public static int GL_TRIANGLE_STRIP = 6;
	final public static int GL_TRIANGLE_FAN = 7;
	public int primitiveType = GL_LINE_STRIP;
	// render mode
	
	final public static int GL_FLAT = 0;
	final public static int GL_SMOOTH = 1;
//	final public static int TEXTURED = 2;
	// kombination
	//      | FLAT | SMOOTH==GOURAUD | TEXTURED
	// NO Z | F    | G               | T
	//    Z | FZ   | GZ              | TZ
	public int shadeModel = GL_FLAT;
	public boolean lighting = false;
	int light_ambient;
	
	Material material = new Material();	// front
	Vector lights = new Vector();		// vector lights
	v eye;	// position of eye in object space
		
	public int lineWidth = 1;
	
	public mat view = new mat();		// modelview matrix
	public mat proj = new mat();		// projective
	public mat viewProj = new mat();
	public v viewPortHalfSize = new v();
	public v viewPortCenter = new v();
	
	// vertex position
	public v first;		// for GL_LINE_LOOP, GL_TRIANGLE_FAN
	public v next0;
	public v next1;
	
	// transformed position
	public vi actual_tr;
	public vi first_tr;
	public vi next0_tr;
	public vi next1_tr;
	
	// vertex color
//	public int actual_c;	// in Device2D
	public int first_c;
	public int next0_c;
	public int next1_c;
	// normal vector
	public v actual_n = new v();
	public v first_n = new v();
	public v next0_n = new v();
	public v next1_n = new v();
	
	public Device3D(){}
	public Device3D(int sizeX_, int sizeY_)
	{
		super(sizeX_, sizeY_);
	}
	
	public vi transform(v a)
	{
		vi o = new vi();
	//	v u = viewProj.mul(a);
		v u = viewProj.mulPerspectiveDivisionPD(a);
		// u ma v suradnice medziach [-1,1]
		o.x = (int)(u.x*viewPortHalfSize.x + viewPortCenter.x);
		o.y = (int)(u.y*viewPortHalfSize.y + viewPortCenter.y);
		return o;
	}
	
	public void drawLineF(v a, v b)
	{
		if(lineWidth==1)
			drawLineF( transform(a), transform(b));
		else
			drawLineF( transform(a), transform(b), lineWidth);
	}
	
	public void drawLineG(v a, v b, int ac, int bc)
	{
		if(lineWidth==1)
			drawLineG( transform(a), transform(b), ac, bc);
		else
			drawLineG( transform(a), transform(b), ac, bc, lineWidth);		
	}
	public void drawTriangleF(v a, v b, v c)
	{
		drawTriangleF( transform(a), transform(b), transform(c));
	}
	
	public void drawTriangleG(v a, v b, v c, int ac, int bc, int cc)
	{
		drawTriangleG( transform(a), transform(b), transform(c), ac, bc, cc);
	}
	
	public void drawTriangleG(v a, v b, v c, v an, v bn, v cn)
	{
		drawTriangleG(transform(a), transform(b), transform(c), 
		              computeLighting(a, an), computeLighting(b, bn), computeLighting(c, cn) );
	}
	
/*	public boolean lighting = false;
	int light_ambient;
	
	Material material = new Material();	// front
	Vector lights = new Vector();		// vector lights*/
	
	public void recomputeLightPositions()
	{
		mat view_inv = view.Inverse();
		eye = mat.mul(view_inv,new v(0,0,0));
		for(int i=0; i<lights.size(); i++)
		{
			Light l = (Light)lights.get(i);
			if(l.isObjectSpace)
			{
				l.position_obj_sp.set(l.position);
			}
			else
			{
				if(l.directional)
					l.position_obj_sp.set(mat.mul3x3(view_inv,l.position));
				else
					l.position_obj_sp.set(mat.mul(view_inv,l.position));
			}
			if(l.directional)
				l.position_obj_sp.Normalize();
		}
	}
	
	public int computeLighting(v pos, v normal)
	{
		if(material==null)
			return 0xffffff;
		int c = material.emission;
		
		// c += light_ambient*material.ambient;
		c = col.add(c,col.mul(light_ambient,material.ambient));
		
		for(int i=0; i<lights.size(); i++)
		{
			int cl=0;
			Light l = (Light)lights.get(i);
			
			// cl = light.ambient*material.ambient;
			cl = col.mul(l.ambient,material.ambient);
			
			if(l.directional)
			{
				// vp = l.pos - pos;
				float brightness = v.DOT3_clamp(l.position_obj_sp,normal);
				// cl += l.diffuse*material.diffuse*brightness;
				cl = col.add(cl,col.mul(col.mul(l.diffuse,material.diffuse),brightness));
				if(brightness!=0.f)
				{
					v ve = v.Normalize(v.sub(eye,pos));
					v h = v.Normalize(v.add(ve,l.position_obj_sp));
					brightness = (float)Math.pow(v.DOT3_clamp(h,normal),material.shininess);
					// cl += l.specular*material.specular*brightness;
					cl = col.add(cl,col.mul(col.mul(l.specular,material.specular),brightness));
				}				
			}
			else // point light
			{
				// vp = l.pos - pos;
				v vp = v.sub(l.position_obj_sp,pos);
				vp.Normalize();
				float brightness = v.DOT3_clamp(vp,normal);
				// cl += l.diffuse*material.diffuse*brightness;
				cl = col.add(cl,col.mul(col.mul(l.diffuse,material.diffuse),brightness));
				if(brightness!=0.f)
				{
					v ve = v.Normalize(v.sub(eye,pos));
					v h = v.Normalize(v.add(ve,vp));
					brightness = (float)Math.pow(v.DOT3_clamp(h,normal),material.shininess);
					// cl += l.specular*material.specular*brightness;
					cl = col.add(cl,col.mul(col.mul(l.specular,material.specular),brightness));
				}
			}
			// attenuation
			c = col.add(c,cl);
		}
		return c;
	}
}
