/*
 * GL.java
 *
 * Created on Nedee, 2004, oktber 24, 19:23
 */

package graphics.device;
import graphics.math.*;
import graphics.device.Device2D;
import graphics.device.Device3D;
/**
 *
 * @author  Marek Mizanin
 * GL - graphics library, all function with gl prefix
 */
public class GL extends Device3D
{
	public GL(){}
	public GL(int sizeX_, int sizeY_)
	{
		super(sizeX_, sizeY_);
		glViewport(0,0,sizeX_,sizeY_);
	}
	
	public void glLoadViewMatrix(mat m)
	{
		view.set(m);
		viewProj.set(mat.mul(proj,view));
	}
	
	public void glLoadProjectionMatrix(mat m)
	{
		proj.set(m);
		viewProj.set(mat.mul(proj,view));
	}
	
	public void gluPerspective( float fovy, float aspect, float zNear, float zFar)
	{
		float xmin, xmax, ymin, ymax;

		ymax = zNear * (float)Math.tan(fovy * Math.PI / 360.0);
		ymin = -ymax;
		xmin = ymin * aspect;
		xmax = ymax * aspect;

		glFrustum(xmin, xmax, ymin, ymax, zNear, zFar);
	}
	public void glFrustum( float left, float right, float bottom, float top, float znear, float zfar)
	{
		float x, y, a, b, c, d;

		x = (2.f * znear) / (right - left);
		y = (2.f * znear) / (top - bottom);
		a = (right + left) / (right - left);
		b = (top + bottom) / (top - bottom);
		c = -(zfar + znear) / ( zfar - znear);
		d = -(2.f * zfar * znear) / (zfar - znear);

		proj.m11 = x;    proj.m12 = 0.0f; proj.m13 = a;     proj.m14 = 0.0f;
		proj.m21 = 0.0f; proj.m22 = y;    proj.m23 = b;     proj.m24 = 0.0f;
		proj.m31 = 0.0f; proj.m32 = 0.0f; proj.m33 = c;     proj.m34 = d;
		proj.m41 = 0.0f; proj.m42 = 0.0f; proj.m43 = -1.0f; proj.m44 = 0.0f;
	}
	
	public void glOrtho(float left, float right, float bottom, float top, float nearZ, float farZ)
	{
	   proj.m11 = 2.0f / (right-left);
	   proj.m12 = 0.0f;
	   proj.m13 = 0.0f;
	   proj.m14 = -(right+left) / (right-left);

	   proj.m21 = 0.0f;
	   proj.m22 = 2.0f / (top-bottom);
	   proj.m23 = 0.0f;
	   proj.m24 = -(top+bottom) / (top-bottom);

	   proj.m31 = 0.0f;
	   proj.m32 = 0.0f;
	   proj.m33 = -2.0f / (farZ-nearZ);
	   proj.m34 = -(farZ+nearZ) / (farZ-nearZ);

	   proj.m41 = 0.0f;
	   proj.m42 = 0.0f;
	   proj.m43 = 0.0f;
	   proj.m44 = 1.0f;
	}
	
	public void glViewport( int x, int y, int width, int height)
	{
		viewPortHalfSize.x = (float)width*0.5f;
		viewPortHalfSize.y =-(float)height*0.5f;
		viewPortCenter.x = (float)(width-x)*0.5f;
		viewPortCenter.y = (float)(height-y)*0.5f;			
	}
	
	public void glVertex3fv(v a)
	{
		switch(primitiveType)
		{
	//	case GL_POINTS:
	//		drawPoint(a);
	//		break;
		case GL_LINES:
			if(shadeModel==GL_SMOOTH)
			{
				if(next0==null)
				{
					next0 = a;
					next0_c = actual_c;
					return;
				}
				drawLineG(next0, a, next0_c, actual_c); 
				next0 = null;
			}
			else  // GL_FTAT
			{
				if(next0==null)
				{
					next0 = a;
					return;
				}
				drawLineF(next0, a);
				next0 = null;
			}
			break;
		case GL_LINE_LOOP:
			if(first==null)
			{
				first = a;
				first_c = actual_c;
			}			
		case GL_LINE_STRIP:
			if(shadeModel==GL_SMOOTH)
			{
				if(next0==null)
				{
					next0 = a;
					next0_c = actual_c;
					return;
				}
				drawLineG(next0, a, next0_c, actual_c); 
				next0 = a;
				next0_c = actual_c;
			}
			else // GL_FLAT
			{
				if(next0==null)
				{
					next0 = a;
					return;
				}
				drawLineF(next0, a); 
				next0 = a;
			}
			break;
		case GL_TRIANGLES:
			if(shadeModel==GL_SMOOTH)
			{
				if(lighting)
				{
					if(next0==null)
					{
						next0 = a;
						next0_n.set(actual_n);
						return;
					}
					if(next1==null)
					{
						next1 = a;
						next1_n.set(actual_n);
						return;
					}
					drawTriangleG(next0, next1, a, next0_n, next1_n, actual_n);
					next0 = null;
					next1 = null;					
				}
				else // no lighting
				{
					if(next0==null)
					{
						next0 = a;
						next0_c = actual_c;
						return;
					}
					if(next1==null)
					{
						next1 = a;
						next1_c = actual_c;
						return;
					}
					drawTriangleG(next0, next1, a, next0_c, next1_c, actual_c);
					next0 = null;
					next1 = null;					
				}
			}
			else // GL_FLAT
			{
				if(lighting)
				{
					if(next0==null)
					{
						next0 = a;
						return;
					}
					if(next1==null)
					{
						next1 = a;
						return;
					}
					actual_c = computeLighting(a, actual_n);
					drawTriangleF(next0, next1, a);
					next0 = null;
					next1 = null;					
				}
				else // no lighting
				{
					if(next0==null)
					{
						next0 = a;
						return;
					}
					if(next1==null)
					{
						next1 = a;
						return;
					}
					drawTriangleF(next0, next1, a);
					next0 = null;
					next1 = null;					
				}				
			}
			break;
		case GL_TRIANGLE_STRIP:
			if(shadeModel==GL_SMOOTH)
			{
				if(lighting)
				{
					int actual_c_;
					if(next0==null)
					{
						next0 = a;
						next0_tr = transform(a);
					//	next0_n.set(actual_n);
						next0_c = computeLighting(a, actual_n);
						return;
					}
					if(next1==null)
					{
						next1 = a;
						next1_tr = transform(a);
					//	next1_n.set(actual_n);
						next1_c = computeLighting(a, actual_n);
						return;
					}
				//	drawTriangleG(next0, next1, a, next0_n, next1_n, actual_n);
					actual_c_ = computeLighting(a, actual_n);
					actual_tr = transform(a);
				//	drawTriangleG(next0, next1, a, next0_c, next1_c, actual_c_);
					drawTriangleG(next0_tr, next1_tr, actual_tr, next0_c, next1_c, actual_c_);
					next0 = next1;
					next0_tr = next1_tr;
					next0_c = next1_c;
				//	next0_n.set(next1_n);
					next1 = a;
					next1_tr = actual_tr;
					next1_c = actual_c_;
				//	next1_n.set(actual_n);
					cullBack = !cullBack;					
				}
				else // no lighting
				{
					if(next0==null)
					{
						next0 = a;
						next0_tr = transform(a);
						next0_c = actual_c;
						return;
					}
					if(next1==null)
					{
						next1 = a;
						next1_tr = transform(a);
						next1_c = actual_c;
						return;
					}
				//	drawTriangleG(next0, next1, a, next0_c, next1_c, actual_c);
					actual_tr = transform(a);
					drawTriangleG(next0_tr, next1_tr, actual_tr, next0_c, next1_c, actual_c);
					next0 = next1;
					next0_tr = next1_tr;
					next0_c = next1_c;
					next1 = a;
					next1_tr = actual_tr;
					next1_c = actual_c;
					cullBack = !cullBack;					
				}
			}
			else // GL_FLAT
			{
				if(lighting)
				{
					if(next0==null)
					{
						next0 = a;
						next0_tr = transform(a);
						return;
					}
					if(next1==null)
					{
						next1 = a;
						next1_tr = transform(a);
						return;
					}
					actual_c = computeLighting(a, actual_n);
					actual_tr = transform(a);
					drawTriangleF(next0_tr, next1_tr, actual_tr);
				//	drawTriangleF(next0, next1, a);
					next0 = next1;
					next0_tr = next1_tr;
					next1 = a;
					next1_tr = actual_tr;
					cullBack = !cullBack;						
				}
				else // no lighting
				{
					if(next0==null)
					{
						next0 = a;
						next0_tr = transform(a);
						return;
					}
					if(next1==null)
					{
						next1 = a;
						next1_tr = transform(a);
						return;
					}
					actual_tr = transform(a);
					drawTriangleF(next0_tr, next1_tr, actual_tr);
				//	drawTriangleF(next0, next1, a);
					next0 = next1;
					next0_tr = next1_tr;
					next1 = a;
					next1_tr = actual_tr;
					cullBack = !cullBack;					
				}				
			}
			break;
		case GL_TRIANGLE_FAN:			
			if(shadeModel==GL_SMOOTH)
			{
				if(lighting)
				{
					int actual_c_;
					if(first==null)
					{
						first = a;
						first_tr = transform(a);
					//	first_n.set(actual_n);
						first_c = computeLighting(a, actual_n);
					}
					if(next0==null)
					{
						next0 = a;
						next0_tr = transform(a);
					//	next0_n.set(actual_n);
						next0_c = computeLighting(a, actual_n);
						return;
					}
					actual_c_ = computeLighting(a, actual_n);
					actual_tr = transform(a);
					drawTriangleG(first_tr, next0_tr, actual_tr, first_c, next0_c, actual_c_);
				//	drawTriangleG(first, next0, a, first_c, next0_c, actual_c_);
				//	drawTriangleG(first, next0, a, first_n, next0_n, actual_n);
					next0 = a;
					next0_tr = actual_tr;
				//	next0_n.set(actual_n);
					next0_c = actual_c_;					
				}
				else // no lighting
				{
					if(first==null)
					{
						first = a;
						first_tr = transform(a);
						first_c = actual_c;
					}
					if(next0==null)
					{
						next0 = a;
						next0_tr = transform(a);
						next0_c = actual_c;
						return;
					}
					actual_tr = transform(a);
					drawTriangleG(first_tr, next0_tr, actual_tr, first_c, next0_c, actual_c);
				//	drawTriangleG(first, next0, a, first_c, next0_c, actual_c);
					next0 = a;
					next0_tr = actual_tr;
					next0_c = actual_c;					
				}
			}
			else // GL_FLAT
			{
				if(lighting)
				{
					if(first==null)
					{
						first = a;
						first_tr = transform(a);
					}
					if(next0==null)
					{
						next0 = a;
						next0_tr = transform(a);
						return;
					}
					actual_c = computeLighting(a, actual_n);
					actual_tr = transform(a);
					drawTriangleF(first_tr, next0_tr, actual_tr);
				//	drawTriangleF(first, next0, a);
					next0 = a;
					next0_tr = actual_tr;
				}
				else // no lighting
				{
					if(first==null)
					{
						first = a;
						first_tr = transform(a);
					}
					if(next0==null)
					{
						next0 = a;
						next0_tr = transform(a);
						return;
					}
					actual_tr = transform(a);
					drawTriangleF(first_tr, next0_tr, actual_tr);
				//	drawTriangleF(first, next0, a);
					next0 = a;
					next0_tr = actual_tr;				
				}				
			}
			break;
		}
	}
	public void glVertex3f(float a, float b, float c)
	{
		glVertex3fv(new v(a,b,c));
	}

	public void glBegin(int type)
	{
		cullBack = cullBackSet;
		primitiveType = type;
		first=null;
		next0=null;
		next1=null;	
	}
	public void glEnd()
	{
		if(primitiveType == GL_LINE_LOOP)
		{
			if(shadeModel==GL_SMOOTH)
				drawLineG(next1, first, next1_c, first_c);
			else
				drawLineF(next1, first);
		}	
	}
	public void glColor3f(float r, float g, float b)
	{
		actual_c = col.make(r,g,b);
	}
	public void glColor3fv(int a)
	{
		actual_c = a;
	}
	public void glColor3fv(v a)
	{
		actual_c = col.make(a.x,a.y,a.z);
	}
	public void glNormal3f(float x, float y, float z)
	{
		actual_n.set(x,y,z);
	}
	public void glNormal3fv(v a)
	{
		actual_n.set(a);
	}
	
	public void glLineWidth(int width)
	{
		if(width<1)width=1;
		lineWidth = width;
	}
	public void glShadeModel(int model)
	{
		shadeModel = model;
	}
	public void glLightModelAmbient(int ambient)
	{
		light_ambient = ambient;
	}
	public void glEnableLighting(){lighting = true;}
	public void glDisableLighting(){lighting = false;}
	public void glMaterial(Material a){ material.set(a);}
	public void glAddLight(Light a)
	{
		lights.add(new Light(a));
	}
	public void glRemoveAllLights()
	{
		lights.clear();
	}
	public void glRecomputeLightPositions()
	{
		recomputeLightPositions();
	}
	public void glEnableCullFace(){ cullFace=true; }
	public void glDisableCullFace(){ cullFace=false; }
	public void glCullFaceBack(){ cullBackSet=true; }
	public void glCullFaceFront(){ cullBackSet=false; }
}
